'''
local exceptions
'''

from requests import HTTPError
from requests.models import Response


# ================================================================================================
class RestClientException(Exception):
    ''' wrapper exception '''
    pass


class RestClientResourceError(RestClientException):
    ''' wrapper exception '''
    pass


class RestClientConfigurationError(RestClientException):
    ''' wrapper exception '''
    pass


class RestClientQueryError(RestClientException):
    ''' wrapper exception '''
    pass


class InvalidTargetError(RestClientException):
    """An error when specifying an invalid target for a given REST API."""
    pass

    # def __init__(self, name, target):
        # """ InvalidTargetError constructor

            # :param name: The name of the REST API client
            # :type name: ``string``

            # :param target: The REST API target name
            # :type target: ``string``

        # """
        # super()"'{target}' is not a valid target for '{name}'".format(
            # target=target,
            # name=name
        # ))


class InvalidResourceError(RestClientException):
    """An error when specifying an invalid resource for a given REST API."""
    def __init__(self, name: str, resource: str):
        """ InvalidResourceError constructor

            :param name: The name of the REST API client
            :param resource: The REST API resource name

        """
        response = f"'{resource}' is not a valid resource for '{name}'"
        super().__init__(response)


class RestResourceMissingContentError(RestClientResourceError):
    ''' wrapper exception '''
    pass


class RestResourceNotFoundError(RestClientResourceError):
    ''' wrapper exception '''
    pass


class RestAccessDeniedError(RestClientResourceError):
    ''' wrapper exception '''
    pass


class RestLoginError(RestClientResourceError):
    ''' wrapper exception '''
    pass

class RestInternalServerError(RestClientResourceError):
    ''' wrapper exception '''
    pass

class RestBadRequestError(RestClientResourceError):
    ''' wrapper excpetion '''
    pass

class RestResourceHTTPError(HTTPError):
    """An error when specifying an invalid target for a given REST API."""
    def __init__(self, response_object, *args, **kwargs):
        """
        RestResourceError constructor
        """

        assert isinstance(response_object, Response)
        self.response = response_object
        self.code = self.response.status_code
        self.reason = self.response.reason
        
        if self.code == 400:
            raise RestBadRequestError('Bad request for resource %s' % (self.response.url,))
        elif self.code == 404:
            raise RestResourceNotFoundError('Object could not be found in database')
        elif self.code in (401, 402, 403):
            raise RestAccessDeniedError('error %d: Access is denied to resource %s' % (self.code, self.response.url))
        elif self.code in (500, ):
            raise RestInternalServerError('error %d: Internal Server error (%s)' % (self.code, self.reason))
        else:
            raise Exception('REST error %d: %s' % (self.code, self.reason))


        super().__init__(*args, **kwargs)



__version__ = '3.0.0'

from .resource import JSONResource
from .conf import APIConfig, ResourceConfig, BodyParameter, QueryParameter
from .exception import  RestClientConfigurationError
from .resource import API




import unittest
from rest_client.utils import URLValidator
from rest_client.exception import RestClientConfigurationError

class TestURLValidator(unittest.TestCase):

	def setUp(self):
		self.validator = URLValidator()

	def tearDown(self):
		pass

	def test_good_urls(self):
		'''
		check if a valid URL is set to valid
		'''
		self.validator.check('http://google.com', require_path=False)
		self.validator.check('https://www.google.com/search?q=test', require_path=False)
		self.validator.check('https://www.google.com/search?q=test', require_path=True)

	def test_bad_urls(self):
		'''
		check if an invalid URL is set to invalid
		'''

		with self.assertRaises(RestClientConfigurationError):
			self.validator.check('http://google.com', require_path=True)

		with self.assertRaises(RestClientConfigurationError):
			self.validator.check('test', require_path=True)



'''
Contains all the configuration classes to create a API Configuration
'''
from collections import defaultdict
from typing import Optional, Type

import logging
logger = logging.getLogger(__name__)

# ================================================================================================
# local imports
from .auth import AuthConfig
from .resource import Resource, JSONResource
from .exception import RestClientConfigurationError
from .utils import URLValidator

# ================================================================================================
#  Interface tweak
from requests.packages.urllib3 import disable_warnings
from requests.packages.urllib3.exceptions import InsecureRequestWarning
disable_warnings(InsecureRequestWarning)

# ================================================================================================
class ParameterConfig:
    '''
    Contain and validate parameters for a REST endpoint. As this is a configuration container only, the main
    purpose is to store the config and check if the input aligns with the intended use. There is no
    validation beyond this point
    '''

    # -----------------------------------------------------------------------------------------------------
    def __init__(self, name: str,
                 required: bool = False,
                 multiple: bool = False,
                 exclusion_group: Optional[str] = None,
                 default: Optional[str] = None,
                 choices: Optional[list] = None,
                 description: Optional[str] = None
                 ):
        '''
        Parameter configuration. Details the name and limitations on the REST parameter and how it
        interacts with other parameters within the same endpoint. 

        :param name: the 'remote' name of the parameter. this name is what the REST resource actually gets to interpret
        :param required: if this parameter is ommitted in the qyery, throw an exception
        :param multiple: if set to True, the value of the query parameter is a list
        :param exclusion_group: parameters in the same exclusion group may not be used together
        :param default: the default entry if this parameter is not supplied
        :param choices: a list of possible values for this parameter
        :param description: any information about the parameter, such as data format
        '''

        self.name = name
        self.required = required
        self.multiple = multiple
        self.exclusion_group = exclusion_group
        self.default = default
        self.choices = choices
        self.description = description or ""
        self._validate()

    # -----------------------------------------------------------------------------------------------------
    def _validate(self):
        '''
        internal routine to check a set of rules to validate if the ParameterConfig is configured correctly
        '''
        if not isinstance(self.required, bool):
            raise RestClientConfigurationError('parameter "required" must be boolean')
        if not isinstance(self.multiple, bool):
            raise RestClientConfigurationError('parameter "multiple" must be boolean')
        if not isinstance(self.description, str):
            raise RestClientConfigurationError('parameter "description" must be string')
        if self.exclusion_group:
            if not isinstance(self.exclusion_group, str):
                raise RestClientConfigurationError("group name must be a string")
            if not self.exclusion_group.strip():
                raise RestClientConfigurationError("group name must be a string")
        if self.default and self.required:
            raise RestClientConfigurationError("you cannot combine required=True and a default setting")
        if self.choices and not isinstance(self.choices, list):
            raise RestClientConfigurationError("choices must be a list")
        if self.default and self.choices:
            if self.default not in self.choices:
                raise RestClientConfigurationError("if there is a choices list, default must be in this list")


# ================================================================================================
class QueryParameter(ParameterConfig):
    '''
    Subclass to specify parameters to be placed in the query part of the REST request
    '''

    call_location = 'query'


# ================================================================================================
class BodyParameter(ParameterConfig):
    '''
    Subclass to specify parameters to be placed in the query part of the REST request
    '''

    call_location = 'body'


# ================================================================================================
class ResourceConfig:
    '''
    contain and validate details for a REST endpoint. Effectively this creates an ORM wrapper around a
    REST endpoint, pretending it is a python object
    
    '''

    # -----------------------------------------------------------------------------------------------------
    def __init__(self,
                 path: list,
                 method: str,
                 parameters: Optional[dict] = None,
                 headers: Optional[dict] = None,
                 processor: Optional[Type[Resource]] = None,
                 description: Optional[str] = None,
                 path_description: Optional[dict] = None
                 ):
        '''
        Constructor, stores externally supplied parameters and validate the quality of it
        
        :param path: a list separation of the path components, e.g. ['api','v2','user','{name},'stats'] where
                     names in brackets are converted to path parameters. 
        :param method: one of GET,PUT,etc
        :param parameters: a dictionary of ParameterConfig instances that each describe one parameter. This is
                    relevant for body and query parameters only, path parameters are specified in the path itself
                    and subsequent annotation of those parameters is done in path_description.
        :param headers: a dictionary of headers that will be provided to the endpoint. Typical use is the response_type
        :param processor: a subclass of rest_client.RestResource that handles specific use cases.
                    This now defaults to JSONResource as this is the most common use type
        :param description: A general description of the endpoint that can be obtained by the user through the description
                    property of the endpointconfig instance
        :param path_description: a dictionary that provides a description for each path parameter.
       
        '''
        self.path = path
        self.description = description
        self.path_description = path_description
        self.method = method
        self.parameters = parameters or {}
        self.headers = headers

        #  we cannot set default processor above in the parameters as this means all endpoints
        #  share the same processor instance, and they cross-contaminate . By setting this below
        #  we enforce recreation of a new unique Resource instance each time
        self.processor = processor or JSONResource()
        self.validate()

    # ----------------------------------------------------
    def validate(self):
        ''' Check quality of each parameter and its type.
        Each parameter is checked for type, and if a specific substructure is required
        then this is also introspected. Currently Method is limited to GET or POST for no reason other then
        no tests were conducted with PUT, HEAD etc etc
        
        :raises RestClientConfigurationError: No response is provided if there is no problem
        
        '''

        # description --------------------
        if self.description and not isinstance(self.description, str):
            raise RestClientConfigurationError("description is not a string")

        if self.path_description and not isinstance(self.path_description, dict):
            raise RestClientConfigurationError("path_description is not a dictionary")

        # path --------------------
        if not isinstance(self.path, list):
            raise RestClientConfigurationError("path is not a list")

        # headers  --------------------
        if self.headers:
            if not isinstance(self.headers, dict):
                raise RestClientConfigurationError("header is not a dict")
        else:
            self.headers = {}

        # method  --------------------
        if self.method not in ['GET', 'POST']:
            raise RestClientConfigurationError("method must be GET or POST")

        #  parameters -------------------------------
        if not isinstance(self.parameters, dict):
            raise RestClientConfigurationError("parameters must be dictionary")
        for key, val in self.parameters.items():
            if not isinstance(val, ParameterConfig):
                raise RestClientConfigurationError("Parameter '%s' must be ParameterConfig instance" % str(key))

        #  resource class ----------------------------------
        if self.processor:
            if not isinstance(self.processor, Resource):
                raise RestClientConfigurationError("processor must be subclass of RestResource")

        # integration tests ---------------------------------------------------
        if self.method == 'GET':
            for key in self.parameters:
                if self.parameters[key].call_location == 'body':
                    raise RestClientConfigurationError('body parameter not allowed in GET request')

    # --------------------------------------------------------------------------------------------
    def apply_default_headers(self, default):
        '''
        For internal use. Update endpoint parameters from a shared default. This allows the user to set
        e.g. headers that are applicable to multiple endpoints in a single activity.
        
        Note that this default only provides functionality for headers
        '''

        # check types
        if not isinstance(default, dict):
            raise RestClientConfigurationError('default must be a dictionary')

        # apply defaults
        def_head = default.copy()
        if self.headers:
            def_head.update(self.headers)
        self.headers = def_head

        # re-validate to be sure current data is OK
        self.validate()

    # ---------------------------------------------------------------------------------------------
    @property
    def path_parameters(self) -> list:
        """ Lists the (always required) path parameters for the specified REST API resource. THis list is obtained
        by checking the path list (['api','v2','{para}','details']) for items that are within curly brackets {}.
        These parameters are stripped and the remainder is added to the path parameter list

        :return: A list of the path parameters for the specified REST API resource
        """

        path_parameters = []
        for part in self.path:
            if part.startswith("{") and part.endswith("}"):
                path_parameters.append(part[1:-1])
        return path_parameters


    # ---------------------------------------------------------------------------------------------
    @property
    def query_parameter_groups(self) -> dict:
        """ Lists the different groups of query parameters for the specified
            REST API resource. When query parameters are in a group, only one of
            them can be used in a query at a time, unless the 'multiple' property
            has been used for every query parameter of that group.

        :return: A dictionary of the different groups (key) of query parameters (value, is list) for the specified REST API resource
        """

        result = defaultdict(list)
        if self.parameters:
            for key, item in self.parameters.items():
                assert isinstance(item, ParameterConfig)
                if item.exclusion_group:
                    result[item.exclusion_group].append(key)
        return dict(result)

    # ---------------------------------------------------------------------------------------------
    @property
    def query_parameters(self) -> dict:
        """ Lists the required and optional query parameters for the specified REST API resource.
            Also summarises the query parameters that can be multiple.

            :return: A dictionary of the 'optional', 'required' and 'multiple' (keys) query parameters
                    (value, a list) for the specified REST API resource. 
        """
        result = {"required": [], "optional": [], "multiple": []}
        for para_name, para_set in self.parameters.items():
            if para_set.required:
                result['required'].append(para_name)
            else:
                result['optional'].append(para_name)

            if para_set.multiple:
                result['multiple'].append(para_name)

        return result

    # --------------------------------------------------------------------------------------------
    @property
    def all_query_parameters(self):
        """ Lists the required and optional query parameters for the specified REST API resource.
            Also summarises the query parameters that can be multiple.

            :return: A list of parameters
            :rtype: ``list``
        """
        params = self.query_parameters
        return params["optional"] + params["required"]

    # --------------------------------------------------------------------------------------------
    @property
    def required_parameters(self):
        """ Lists the required parameters for the specified REST API resource.
            Also summarises the query parameters that can be multiple.

            :return: A dictionary of the 'optional', 'required' and 'multiple' (keys) query parameters (value, a list) for the specified REST API resource
            :rtype: ``list``
        """
        return self.path_parameters + self.query_parameters["required"]

    # --------------------------------------------------------------------------------------------
    @property
    def multiple_parameters(self):
        """ Returns all parameters that can be used simultaneously

            :return: A list of parameters
            :rtype: ``list``
        """
        return self.query_parameters["multiple"]


    # --------------------------------------------------------------------------------------------
    @property
    def all_parameters(self):
        """ Aggregates all parameters into a single structure

            :return: A list of parameters
            :rtype: ``list``
        """
        all_parameters = self.all_query_parameters + self.path_parameters
        return all_parameters

    # ---------------------------------------------------------------------------------------------
    @property
    def as_dict(self) -> dict:
        """ show all parameters in path or query

            :return: A dictionary that contains required and optional parameters.
        """

        result = {"required": [], "optional": []}
        result["required"].extend(self.path_parameters)

        qp = self.query_parameters
        result["required"].extend(qp['required'])
        result["optional"].extend(qp["optional"])

        return result


    # ---------------------------------------------------------------------------------------------
    @property
    def defaults(self):
        """
        Returns a dict with all default parameters and their value
        :return: A dictionary
        :rtype: ``dict``
        """
        defaults = {x:y.default for x, y in self.parameters.items() if y.default}
        return defaults


#==================================================================================================
class APIConfig:
    '''
    Class to configure and validate endpoints
    '''

    authentication = None
    url = None
    verify_ssl = False

    def __init__(self):
        self.endpoints = self.get_list_of_endpoints()
        self.apply_defaults()
        self.validate()


    def apply_defaults(self):
        '''
        rotate throught the endpoints and apply the fedault settings
        '''
        if 'default_headers' in dir(self):
            for endpoint in self.endpoints.values():
                endpoint.apply_default_headers(self.default_headers)


    def validate(self):
        """
        Validates a resources configuration and raises appropriate exceptions
        """
        for resource_name, resource_config in self.endpoints.items():
            if resource_name == 'data':
                raise RestClientConfigurationError("resource name may not be named 'data'")

        # check url definition
        if not self.url:
            raise RestClientConfigurationError('url is not set')
        URLValidator().check(self.url, require_path=False)

        if not isinstance(self.verify_ssl, bool):
            raise RestClientConfigurationError('verify_ssl is not True or False')
        
        # optional auth module
        if self.authentication and not isinstance(self.authentication, AuthConfig):
            raise RestClientConfigurationError('authentication attribute is not an initiated instance of AuthConfig')


    def get_list_of_endpoints(self):
        '''
        returns a dictionary of all the defined endpoints
        '''
        endpoints_dict = {}
        dirlist = [x for x in dir(self) if not x.startswith('_')]
        for item in dirlist:
            endpoint = getattr(self, item)
            if endpoint.__class__ == ResourceConfig:
                endpoints_dict[item] = endpoint
        if not endpoints_dict:
            raise RestClientConfigurationError('no endpoints defined for this REST client at all!')
        return endpoints_dict
        #return []





''' Contains a set of related and unrelated functions and classes used elsewhere in this module
'''

import logging
from urllib.parse import urlparse
from .exception import RestClientConfigurationError

logger = logging.getLogger(__name__)

# ###############################################################
class URLValidator:
	"""
	lightweight URL validation checker. This is a minimal implementation compared to e.g. the Django validation
	checker, as the configs are mostly supposed to be fixed in advance of usage.
	"""

	def __init__(self, schemes=None):
		"""
		set the restrictions imposed on the URL to be checked
		"""
		pass

	# ------------------------------------------------------------------
	def check(self, url, require_path=True):
		"""
		check if the presented URL is valid
		"""

		final_url = urlparse(url)
		if not all([final_url.scheme, final_url.netloc]):
			raise RestClientConfigurationError(f'the URL {url} is not valid')
		if len(final_url.netloc.split(".")) == 1 and final_url.netloc != 'localhost':
			raise RestClientConfigurationError(f'the URL {url} is has no domain indication')
		if require_path and not final_url.path:
			raise RestClientConfigurationError(f'the URL {url} has no valid path')
    
	
	




import os
import logging
import requests

from ..exception import RestLoginError, RestClientConfigurationError
from . import NetRCAuth, RESTAuthentication, AuthConfig
from ..utils import URLValidator


logger = logging.getLogger(__name__)


class CASLoginError(RestLoginError):
	pass

class CASGrantingTicketError(CASLoginError):
	pass

class CASServiceTicketError(CASLoginError):
	pass



# ==========================================================================================

class CASAuth(RESTAuthentication):
	'''
	Subclass of the RESTAuthentication that is able to use CAS to request and store
	a ticket-granting-ticket and subsequent service tickets.
	'''

	# -------------------------------------------------------------------------------------
	def __init__(self, rest_client, auth_config_object):
		"""
		CASAuth constructor

		:param rest_client: A reference to the RESTclient object
		:type rest_client: ``RESTclient``

		:param auth_config_object: The configuration object
		:type auth_config_object: ``AuthConfig``

		"""

		super(CASAuth, self).__init__(rest_client)

		config = auth_config_object

		# the remote URL for the CAS server
		self.server = None
		self.service = config.service_name
		self.ticket_path = '/'.join(config.path)  # the URL path

		self.tgt_volatile_storage = False
		self.tgt_file_name = None
		self.__ticket_granting_ticket = None  # content of the TGT


	# -------------------------------------------------------------------------------------
	def login(self, server_url, service_name=None, verify_ssl=False,
	          netrc_path=os.path.expanduser("~/.netrc"),
	          username=None, password=None,
	          tgt_volatile_storage=False,
	          granting_ticket_filepath=None,
	          ticket_granting_ticket=None
	          ):
		"""
		CAS-auth requires a secondary auth, so it has two authentication levels...
		Allow logins via either netRC or user/pass.
		If netrc is true, try this. Else, try user/pass if supplied.
		If neither is available, raise error


		:param server_url: the URL of the CAS server
		:type server_url: ``string_type``

		:param service_name: the name of the service of CAS
		:type service_name: ``string_type``


		:param verify_ssl: Whether the CAS client should verify SSL certificates upon making requests
		:type verify_ssl: ``bool``  or `` string_type`` 

		:param netrc_path: The path to Netrc
		:type netrc_path: ``string_type_or_none``

		:param username: The user for authenticating with the CAS end-point
		:type username: ``string_type_or_none``

		:param password: The password for authenticating with the CAS end-point
		:type password: ``string_type_or_none``

		:param granting_ticket_filepath: The local path where a "ticket getting ticket" will be saved
		:type granting_ticket_filepath: ``string_type``

		:param tgt_volatile_storage: Is the TGT stored on the filesystem or handed over to the user
		:type tgt_volatile_storage: ``boolean``
		"""

		# the connection parameters
		self.server = server_url
		self.service_name = service_name
		self.verify_ssl = verify_ssl
		parent_auth = NetRCAuth(rest_client=self.rest_client)

		# ---- parameter verification -----
		# TGT in file or attribute
		self.tgt_volatile_storage = tgt_volatile_storage
		if tgt_volatile_storage:
			if not ticket_granting_ticket:
				self.ticket_granting_ticket = None
			try:
				self.ticket_granting_ticket = ticket_granting_ticket
			except CASGrantingTicketError:
				# the TGT was not a valid URL: reset local to None to clear out
				logger.debug("[CAS] was provided with unusable TGT: '%s', resettign to None" % ticket_granting_ticket)
				self.ticket_granting_ticket = None

		else:
			if not granting_ticket_filepath:
				raise RestClientConfigurationError('TGT path must be set if volatile_storage is False')
			self.tgt_file_name = os.path.expanduser(granting_ticket_filepath)

		# if username/pass are provided, then always request a new TGT
		if self.are_valid_credentials(username, password):
			self.ticket_granting_ticket = self.request_new_tgt(username, password)
		elif netrc_path:
			# try to get creds from parent, in case we may need it later
			# if this request fails, its ok for now
			try:
				parent_auth.login(netrc_path=netrc_path)
				username = parent_auth.username
				password = parent_auth.password
			except ValueError:
				pass

		# if no TGT, then a new request needs to be made: either using user/pass, or netrc as backup
		if not self.ticket_granting_ticket:
			if not self.are_valid_credentials(username, password):
				raise RestLoginError('no username or password is provided via parameters or in netrc file')
			self.ticket_granting_ticket = self.request_new_tgt(username, password)

		# ok, we should have a TGT, but it may be outdated or otherwise bad
		try:
			self.request_new_service_ticket()
		except CASServiceTicketError as e:
			# we could not get a service ticket, lets try a new tgt
			logger.debug("[CAS] could not get service ticket with old TGT, try to get a new one")
			if not self.are_valid_credentials(username, password):  # this uses the netrc creds, but only after trying without first
				raise RestLoginError('[CAS] could not get service ticket, and netrc credentials are invalid or absent. Exact error msg="%s"' % str(e))
			try:
				self.ticket_granting_ticket = self.request_new_tgt(username, password)
				self.request_new_service_ticket()
			except CASGrantingTicketError as e2:
				raise RestLoginError('[CAS] could not get TGT while using netrc credentials. Exact error msg="%s"' % str(e2))
			except CASServiceTicketError as e2:
				raise RestLoginError('[CAS] could not get service ticket, despite getting a new TGT. Exact error msg="%s"' % str(e2))

		self.is_logged_in = True

	# -------------------------------------------------------------------------------------
	def request_new_service_ticket(self):
		""" Retrieves the service ticket that will ultimately be used inside the request to the REST end-point.

		    :return: The text or token to be used inside the Authorization header of the RESTful request
		    :rtype: ``string_type``
		"""
		logger.debug("[CAS] Requesting new service ticket")

		# allow override of service name
		if self.service_name:
			body = {"service": self.service_name}
		else:
			body = {"service": self.service}

		if not self.ticket_granting_ticket:
			logger.debug("[CAS] No granting ticket available while asking for a service ticket")
			raise CASServiceTicketError('[CAS] No granting ticket available')

		response = requests.post(url=self.ticket_granting_ticket, data=body, verify=self.verify_ssl)
		if not response.ok:
			logger.debug("[CAS] Service ticket request failed")
			raise CASServiceTicketError("Cannot authenticate against CAS service using service name '{service}'. HTTP status code: '{status}'".format(status=response.status_code, service=body['service']))

		return response.text


	# -------------------------------------------------------------------------------------
	@property
	def ticket_granting_ticket(self):
		"""
		Reads the ticket getting ticket that will ultimately be used inside
		the request to the CAS end-point for retrieving a service ticket.
		The "tgtPath" parameter has to point to an existing location.

		:return: The URL to be used for retrieving a service ticket from the CAS end-point
		:rtype: ``string_type``
		"""
		if self.tgt_volatile_storage:
			if not self.__ticket_granting_ticket:
				return None
			logger.debug("[CAS] Reading TGT from memory: contents are '%s'", self.__ticket_granting_ticket)  # printing credentials here!
			return self.__ticket_granting_ticket
		else:
			if not self.tgt_file_name:
				logger.warning('[CAS] no tgt file path provided')
				return None
			elif not os.path.isfile(self.tgt_file_name):
				logger.warning("[CAS] File '%s' does not exist.", self.tgt_file_name)
				dirname =  os.path.dirname(self.tgt_file_name)
				if not os.path.isdir(dirname):
					os.mkdir(dirname, 0o700)
				return None
			else:
				with open(self.tgt_file_name, 'r') as tgt_file:
					tgt = tgt_file.read().strip()
				logger.debug("[CAS] Reading TGT from file: contents are '%s'", tgt)  # printing credentials here!
				if tgt == "":
					os.remove(self.tgt_file_name)
					raise CASGrantingTicketError("[CAS] TGT file at '%s' was empty and has been removed."% self.tgt_file_name)
				else:
					return tgt

	# -------------------------------------------------------------------------------------
	@ticket_granting_ticket.setter
	def ticket_granting_ticket(self, tgt):
		'''
		depending on method, store TGT in file or within class instance
		'''

		# special case: if TGT is None then clear the underlying variable
		if tgt == None:
			self.__ticket_granting_ticket = None
			return

		# Validate the REST API base URL
		url_validator = URLValidator()
		url_validator.check(tgt)

		# the content of tgt is in a variable
		if self.tgt_volatile_storage:
			self.__ticket_granting_ticket = tgt
		else:
			# tgt is on file on disk
			logger.debug("[CAS] TGT URI is '%s'", tgt)
			tgt_dir = os.path.dirname(self.tgt_file_name)
			if not os.path.isdir(tgt_dir):
				os.makedirs(tgt_dir)
			if os.path.isfile(self.tgt_file_name):
				os.remove(self.tgt_file_name)
			with open(self.tgt_file_name, "w") as tgt_file:
				tgt_file.write(tgt)


	# -------------------------------------------------------------------------------------
	def request_new_tgt(self, username, password):
		""" Retrieves the ticket getting ticket that will ultimately be used inside the request to the CAS
		    end-point for retrieving a service ticket.
		    If the "tgtPath" parameter isn't pointing to an existing location, that location will be created.
		    If the file at "tgtPath" exists, it will be replaced by the newly retrieved TGT.
		"""

		if not self.are_valid_credentials(username, password):
			raise CASGrantingTicketError('TGT: no valid username or password available to request a ticket-granting-ticket')

		logger.debug("[CAS] Renewing granting ticket")
		ticket_url = "{server}/{path}".format(server=self.server, path=self.ticket_path)
		response = requests.post(url=ticket_url,
		                         data={"username": username, "password": password},
		                                 verify=self.verify_ssl)

		if response.status_code == 401:
			raise CASGrantingTicketError('TGT: could not login using provided username "%s" and password' % username)
		elif (response.status_code < 200) or (response.status_code >= 300):
			raise CASGrantingTicketError("TGT: Cannot authenticate against CAS using provided 'username:{username}' and 'password'. HTTP status code: '{status}'".format(status=response.status_code, username=username))

		tgt = response.headers["location"]
		return tgt

	## -------------------------------------------------------------------------------------
	def __call__(self, r):
		"""
		Is called by the requests library when authentication is needed while issuing a RESTful request.
		Will retrieve a new service ticket or token if necessary.
		Adds the Authorization header and supplies the correct service ticket or token to it.
		"""
		service_ticket = self.request_new_service_ticket()
		logger.debug("[CAS] add service ticket to request header")
		r.headers['Authorization'] = "CAS {service_ticket}".format(service_ticket=service_ticket)
		return r

# ==========================================================================================
class CasAuthConfig(AuthConfig):
	'''
	CAS authentication specific for the CLS implementation below
	'''

	authentication_module = CASAuth

	# -------------------------------------------------------------------------------------
	def __init__(self, path, service_name):
		'''
		:param path: The absolute path for the ticket granting tickets
		:type path: ``list``

		:param service: The service name used to authenticate with the CAS end-point
		:type service: ``string_type``

		'''

		self.path = path
		self.service_name = service_name



'''
This module contains the Response objects. A response object is a wrapper around the data the comes from the
REST server after the client asks for it. 
'''

import copy
import requests
import logging
from abc import ABC, abstractmethod
from typing import Optional, Type

from requests.packages.urllib3 import disable_warnings
from requests.packages.urllib3.exceptions import InsecureRequestWarning

# ================================================================================================
# local imports
from .exception import RestResourceMissingContentError, RestClientConfigurationError

disable_warnings(InsecureRequestWarning)
logger = logging.getLogger(__name__)

# =================================================================================================

class Response(ABC):

	'''
	Abstract Base class wrapper around the REST response. This is meant to process the returned data
	obtained from the REST request into a python object
	'''

	_response = None
	headers = None
	raw = None
	options = None

	@abstractmethod
	def __init__(self):
		""" RestResponse constructor
		This should be overridden by subclasses so that required parameters are visible through introspection
		"""


	def __call__(self, response: Type[requests.models.Response]):
		""" RestResponse wrapper call
		    :param response: The Requests Response object
		"""
		if not self.options:
			# raise RestClientConfigurationError('configuration is not set for API Response')
			logger.warn('No options are provided')
		
		if not isinstance(response, requests.models.Response):
			raise TypeError('RestResponse expects a requests.models.Response as input')

		self._response = response
		self.headers = response.headers
		self.raw = response.content

		#prepare the content to a python object
		self._parse()
		return self

	def fetch(self):
		'''
		systematic function to indicate that the required content is to be delivered
		'''
		return self.data


	@abstractmethod
	def _check_content(self):
		pass


	@abstractmethod
	def _parse(self):
		pass



#  =========================================================================================================
class JSONResponse(Response):

	def __init__(self, extract_section: Optional[list] = None, create_attribute: Optional[str] = 'results'):
		'''
		Special Wrapper to handle JSON responses. It takes the response object and creates a dictionary.
		Optionally it allows extraction of a payload subsection to a user-defined attribute. In all cases the
		raw data goes to the 'data' property'.
		
		:param extract_section: This indicates which part of the obtained JSON response contains the main
		    payload that should be extracted. The tree is provided as a list of items to traverse
		:param create_attribute: The "results_name" which is the property that will be generated to contain
		    the previously obtained subsection of the json tree
		
		'''

		if extract_section and not isinstance(extract_section, list):
			raise RestClientConfigurationError("extract_section option is not a list")
		self.extract_section = extract_section
		self.create_attribute = create_attribute

	def _check_content(self):
		content_type = self.headers.get('content-type', 'unknown')
		if not "json" in content_type:
			raise TypeError(f'the REST response did not give a JSON but a {content_type}')

	def _parse(self):
		""" Returns the JSON contained in the Requests Response object, following the options specified in the JSON configuration

		    :return: A dictionary containing the Requests Response object, adapted to the JSON configuration
		    :rtype: ``dict``
		"""

		# replace content by decoded content
		self.raw = copy.deepcopy(self._response.json())

		# subset the response dictionary
		json = copy.deepcopy(self._response.json())
		if isinstance(json, dict) and self.extract_section:
			for element in self.extract_section:
				if element in json:
					json = json[element]
				else:
					raise RestResourceMissingContentError(f"Element {element} could not be found")
			setattr(self, self.create_attribute, json)
		self.data = json


#  =========================================================================================================
class CSVRestResponse(Response):

	def _check_content(self):
		content_type = self.headers.get('content-type', 'unknown')
		if not 'text/csv' in content_type:
			raise TypeError(f'the REST response did not give a CSV but a {content_type}')

	def _parse(self) -> list:
		''' processes a raw CSV into lines. For very large content this may be better served by a generator
		'''
		
		data = self._response.content
		self.raw = data.decode('UTF-8')
		data = self.raw.strip().split('\n')
		return [x.split(",") for x in data]



'''
This module handle authentication against a variety of REST services
'''


import os
import logging
from typing import Optional
from netrc import netrc
from urllib.parse import urlparse
import requests

from abc import ABC, abstractmethod

# ================================================================================================
# local imports
from ..exception import RestLoginError

logger = logging.getLogger(__name__)

# ==========================================================================================
class RESTAuthentication(ABC, requests.auth.HTTPBasicAuth):
    '''
    Standard authentication methods and credentials
    This basic authentication does not complain if user is not logging in
    explicitely. This is the most silent form of authentication
    '''

    is_logged_in = False
    username = None
    password = None

    def __init__(self, rest_client, auth_config_object=None):
        """
        basic auth that uses user/pass or netrc
        this is a subclass of HTTPBasicAuth, but differs in that it requires a specific call to
        login() to allow netrc and embedding.

        :param rest_client: A reference to the RESTclient object
        :type rest_client: ``RESTclient``

        """

        self.rest_client = rest_client
        self.auth_config_object = auth_config_object

    @property
    def login_tuple(self):
        '''
        returns a username/password tuple to use in basic authentication
        '''
        return (self.username, self.password)

    @staticmethod
    def is_valid_credential(credential):
        """ Determines whether the given credential is valid.
            The credential should not be None and should not be the empty string.

            :param credential: The credential to check
            :type credential: ``string_type_or_none``
        """
        return credential is not None and credential.strip() != ""

    def are_valid_credentials(self, username, password):
        """ Determines whether given credentials are valid.
            Both username and password should not be None and should not be the empty string.

            :param username: The username to check
            :type username: ``string_type_or_none``

            :param password: The password to check
            :type password: ``string_type_or_none``
        """

        if self.is_valid_credential(username):
            return True
        else:
            if self.is_valid_credential(password):
                raise RestLoginError('provided password but not a username')
            else:
                return False

    # -------------------------------------------------------------------------------
    @abstractmethod
    def login(self):
        '''
        placeholder to enforce subclasses to customize the login method
        '''
        raise NotImplementedError('Define login procedure in subclass')


# ==========================================================================================
class UserPassAuth(RESTAuthentication):
    '''
    Authentication that enforces username / password
    '''

    is_logged_in = False

    # -------------------------------------------------------------------------------
    def login(self, username=None, password=None):
        """ Explicitly do a login with supplied username and password.
            If username and password are supplied, it will use those to login.
            If either or both username and/or password is missing, will try to retrieve the credentials from the netrc file.

            :param username: The user for authenticating with the CAS end-point
            :type username: ``string_type_or_none``

            :param password: The password for authenticating with the CAS end-point
            :type password: ``string_type_or_none``
        """

        if not self.are_valid_credentials(username, password):
            raise ValueError("No or empty username and/or password supplied")
        else:
            self.username = username
            self.password = password

        self.is_logged_in = True


# ==========================================================================================
class NetRCAuth(RESTAuthentication):
    '''
    Authentication that enforces username / password
    '''

    netrc_path = os.path.expanduser("~/.netrc")
    is_logged_in = False

    # -------------------------------------------------------------------------------
    #def __init__(self, rest_client, config):
        #super(UserPassAuth, self).__init__(rest_client)


    # -------------------------------------------------------------------------------
    def login(self, netrc_path: Optional[str] = os.path.expanduser("~/.netrc")):
        """ Explicitly do a login with supplied username and password.
            If username and password are supplied, it will use those to login.
            If either or both username and/or password is missing, will try to retrieve the credentials from the netrc file.

            :param netrc_path: The network path to the netrc file
        """

        if netrc_path:
            logger.debug("Retrieving using netrc.")
            try:
                self.netrc_path = os.path.expanduser(netrc_path)
            except AttributeError as e:
                raise ValueError('could not expand netrc-path. error is "%s"' % str(e))
            nrc = netrc(file=self.netrc_path)
            host = urlparse(self.rest_client.url).hostname
            try:
                (netrc_login, _, netrc_password) = nrc.authenticators(host)
            except TypeError:
                raise ValueError("No credentials found for host '{host}' or 'default' in the netrc file at location '{location}'".format(host=host, location=self.netrc_path))
            if self.are_valid_credentials(netrc_login, netrc_password):
                self.username = netrc_login
                self.password = netrc_password
            else:
                raise ValueError("No valid credentials found for host '{host}' or 'default' in the netrc file at location '{location}'".format(host=host, location=self.netrc_path))
        else:
            raise ValueError("No Netrc path supplied")


# ==========================================================================================
class UserPassOrNetRCAuth(RESTAuthentication):
    '''
    wrapper to allow both netRC and User+Pass logins
    '''

    def login(self, netrc_path: Optional[str] = os.path.expanduser("~/.netrc"), username: Optional[str] = None, password: Optional[str] = None):
        """
        Allow logins via either netRC or user/pass.
        If netrc is true, try this. Else, try user/pass if supplied.
        If neither is available, raise error

        :param netrc_path: The path to the Netrc file
        :param username: The user for authenticating with the CAS end-point
        :param password: The password for authenticating with the CAS end-point
        """

        if username and password:
            parent_auth = UserPassAuth(rest_client=self.rest_client)
            parent_auth.login(username=username, password=password)
        elif netrc_path:
            parent_auth = NetRCAuth(rest_client=self.rest_client, auth_config_object=self.auth_config_object)
            parent_auth.login(netrc_path=netrc_path)
        else:
            raise RestLoginError('not enough data is provided to login')
        self.username = parent_auth.username
        self.password = parent_auth.password

        self.is_logged_in = True




# ==========================================================================================
class AuthConfig(object):
    '''
    Configuration and validation for custom authentication schemas
    '''
    pass


# ==========================================================================================
class UserPassAuthConfig(AuthConfig):
    '''
    Allow authentication via NetRC or User/Password
    '''

    authentication_module = UserPassAuth


# ==========================================================================================
class NetrcOrUserPassAuthConfig(AuthConfig):
    '''
    Allow authentication via NetRC or User/Password
    '''

    authentication_module = UserPassOrNetRCAuth





import unittest

import rest_client
from rest_client import APIConfig, RestClientConfigurationError
from rest_client import QueryParameter, BodyParameter, ResourceConfig
from rest_client import JSONResource
from rest_client.auth import UserPassAuthConfig

class TestMinimal(unittest.TestCase):

	def test_minimal_config(self):
		'''
		check method or path
		'''
		
		class Config(APIConfig):
			url = 'http://localhost'
			ep = ResourceConfig(path=[''], method='GET')
		Config()

		class Config(APIConfig):
			url = 'http://localhost'
			ep = ResourceConfig(path=['test'], method='GET')
		Config()
	
		with self.assertRaises(RestClientConfigurationError):
			class Config(APIConfig):
				url = 'http://localhost'
				ep = ResourceConfig(path=['test'], method='ERR')
			Config()

		with self.assertRaises(RestClientConfigurationError):
			class Config(APIConfig):
				url = 'http://localhost'
				ep = ResourceConfig(path=3, method='GET')
			Config()

		with self.assertRaises(RestClientConfigurationError):
			class Config(APIConfig):
				ep = ResourceConfig(path=[''], method='GET')
			Config()


class TestParameters(unittest.TestCase):

	UrlApiConfig = None

	def setUp(self):
		class UrlApiConfig(APIConfig):
			url = "http://localhost"
		self.UrlApiConfig = UrlApiConfig


	def tearDown(self):
		pass

	def test_good_query_parameters(self):
		# --------------------------------------------------------------
		class Config(self.UrlApiConfig):
			ep = ResourceConfig(path=[''], method='GET',
						  parameters={
							  'para': QueryParameter(name="sort",
													 default='some_default',
													 description='description',
                                ),
                            }
                 )
		Config()

		# --------------------------------------------------------------
		class Config(self.UrlApiConfig):
			ep = ResourceConfig(path=[''], method='GET',
						  parameters={
							  'para': QueryParameter(name="sort",
													 required=False,
                                ),
                            }
                 )
		conf = Config()


	def test_bad_query_parameters(self):
		# --------------------------------------------------------------
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				ep = ResourceConfig(path=[''], method='GET',
							  parameters={
								  'para': QueryParameter(name="sort",
														 required=3
                                 ),
                                    }
                  )
		# --------------------------------------------------------------
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				ep = ResourceConfig(path=[''], method='GET',
							  parameters={
								  'para': QueryParameter(name="sort",
														 multiple=3
                                 ),
                                    }
                  )
		# --------------------------------------------------------------
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				ep = ResourceConfig(path=[''], method='GET',
							  parameters={
								  'para': QueryParameter(name="sort",
														 exclusion_group=3
                                 ),
                                    }
                  )
		# --------------------------------------------------------------
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				ep = ResourceConfig(path=[''], method='GET',
							  parameters={
								  'para': QueryParameter(name="sort",
														 default='some_default',
														 required=True
									),
								}
					 )

	def test_exclusion_group(self):
		# --------------------------------------------------------------
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				ep = ResourceConfig(path=[''], method='GET',
							  parameters={
								  'para': QueryParameter(name="sort",
														 exclusion_group='\n',
                                 ),
                                    }
                  )

		# --------------------------------------------------------------
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				ep = ResourceConfig(path=[''], method='GET',
							  parameters={
								  'para': QueryParameter(name="sort",
														 default='x',
														 choices=['a', 'b'],
                                                                  ),
                                    }
                  )

	# --------------------------------------------------------------
	def test_choices(self):
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				ep = ResourceConfig(path=[''], method='GET',
							  parameters={
								  'para': QueryParameter(name="sort",
														 choices='some_default',
									),
								}
					 )

	# --------------------------------------------------------------
	def test_descriptions(self):
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				ep = ResourceConfig(path=[''], method='GET',
							  parameters={
								  'para': QueryParameter(name="sort",
														 description=3,
                                 ),
                                    }
                  )

	# --------------------------------------------------------------
	def test_bad_body_parameters(self):
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				ep = ResourceConfig(path=[''], method='GET',
							  parameters={
								  'para': BodyParameter(name="sort"),
                                    }
                  )
		
class TestEndpoint(unittest.TestCase):

	UrlApiConfig = None
	
	def setUp(self):
		class UrlApiConfig(APIConfig):
			url = "http://localhost"
		self.UrlApiConfig = UrlApiConfig

	def tearDown(self):
		pass

	# --------------------------------------------------------------
	def test_bad_init_parameters(self):
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				ep = ResourceConfig(path=[''], method='GET',
							  description=3
                  )

		# --------------------------------------------------------------
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				ep = ResourceConfig(path=[''], method='GET',
							  path_description=3
                  )
		# --------------------------------------------------------------
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				ep = ResourceConfig(path=[''], method='GET',
							  headers='none'
                        )
	
	# --------------------------------------------------------------
	def test_no_endpoints(self):
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				default_headers = {'x': 'y', }
			c = Config()
			c.get_list_of_endpoints()

	def test_bad_rest_parameters(self):
		# --------------------------------------------------------------
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				ep = ResourceConfig(path=[''], method='GET',
							  parameters='x'
                  )
		# --------------------------------------------------------------
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				ep = ResourceConfig(path=[''], method='GET',
							  parameters='x'
                  )

		# --------------------------------------------------------------
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				ep = ResourceConfig(path=[''], method='GET',
							  parameters={'x': 'y', }
                  )

	def test_bad_endpoints(self):
		# --------------------------------------------------------------
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				data = ResourceConfig(path=[''], method='GET')
			Config()


	def test_bad_defaults(self):
		# --------------------------------------------------------------
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				ep = ResourceConfig(path=[''], method='GET')
				default_headers = 'error'
			Config()

		# --------------------------------------------------------------
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				ep = ResourceConfig(path=[''], method='GET')
				default_headers = {'key': 3},
			Config()

		# --------------------------------------------------------------
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				ep = ResourceConfig(path=[''], method='GET')
				default_headers = {'key': None},
			Config()


	def test_good_defaults(self):
		# --------------------------------------------------------------
		class Config(self.UrlApiConfig):
			ep = ResourceConfig(path=[''], method='GET')
			default_headers = {'key': 'val'}
		Config()

		# --------------------------------------------------------------
		class Config(self.UrlApiConfig):
			ep = ResourceConfig(path=[''], method='GET', headers={'key': 'oldval', })
			default_headers = {'key': 'val'}
		conf = Config()
		self.assertDictEqual(conf.endpoints['ep'].headers, {'key': 'oldval'})

		# --------------------------------------------------------------
		class Config(self.UrlApiConfig):
			ep = ResourceConfig(path=[''], method='GET', headers={'otherkey': 'val', })
			default_headers = {'key': 'val'}
		conf = Config()
		self.assertDictEqual(conf.endpoints['ep'].headers, {'key': 'val', 'otherkey': 'val'})
		

	def test_descriptions(self):
		# --------------------------------------------------------------

		class Config(self.UrlApiConfig):
			ep = ResourceConfig(path=[''], method='GET',
						  description="OK")
		conf = Config()

		# --------------------------------------------------------------
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				ep = ResourceConfig(path=[''], method='GET',
							  description=3)
			Config()

			# --------------------------------------------------------------
		class Config(self.UrlApiConfig):
			ep = ResourceConfig(path=[''], method='GET',
						  path_description={'x': 'y', })
			Config()

		# --------------------------------------------------------------
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				ep = ResourceConfig(path=[''], method='GET',
							  path_description=3)
			Config()

		

	def test_introspection(self):
		# --------------------------------------------------------------
		class Config(self.UrlApiConfig):
			ep1 = ResourceConfig(path=[''], method='GET',
						   parameters={'p1': QueryParameter('p1'),
									   'p2': QueryParameter('p2', multiple=True),
									   'p3': QueryParameter('p3', required=True),
									   'p4': QueryParameter('p4', default='def'),
                     }
                  )

		c = Config()
		expected = {'required': ['p3'], 'optional': ['p1', 'p2', 'p4'], 'multiple': ['p2']}
		self.assertDictEqual(c.ep1.query_parameters, expected)

		self.assertListEqual(c.ep1.all_query_parameters, ['p1', 'p2', 'p4', 'p3'])
		self.assertListEqual(c.ep1.required_parameters, ['p3'])
		self.assertListEqual(c.ep1.multiple_parameters, ['p2'])
		self.assertListEqual(c.ep1.all_parameters, ['p1', 'p2', 'p4', 'p3'])

		self.assertDictEqual(c.ep1.as_dict, {'required': ['p3'], 'optional': ['p1', 'p2', 'p4']})
		self.assertDictEqual(c.ep1.defaults, {'p4': 'def'})

		# --------------------------------------------------------------
		class Config(self.UrlApiConfig):
			ep1 = ResourceConfig(path=[''], method='GET',
						   parameters={'p1': QueryParameter('p1'),
									   'p2': QueryParameter('p2', multiple=True, exclusion_group='a'),
									   'p3': QueryParameter('p3', required=True, exclusion_group='a'),
									   'p4': QueryParameter('p4'),
                     }
                        )

		c = Config()
		expected = {'a': ['p2', 'p3']}
		self.assertDictEqual(c.ep1.query_parameter_groups, expected)
		
		# --------------------------------------------------------------
		class Config(self.UrlApiConfig):
			ep1 = ResourceConfig(path=['x', 'y', 'z'], method='GET')

		c = Config()
		expected = []
		self.assertListEqual(expected, c.ep1.path_parameters)

		# --------------------------------------------------------------
		class Config(self.UrlApiConfig):
			ep1 = ResourceConfig(path=['x', '{y}', 'z'], method='GET')

		c = Config()
		expected = ['y']
		self.assertListEqual(expected, c.ep1.path_parameters)

class TestAuthentication(unittest.TestCase):

	UrlApiConfig = None

	def setUp(self):
		class UrlApiConfig(APIConfig):
			url = "http://localhost"
			ep = ResourceConfig(path=[''], method='GET')
		self.UrlApiConfig = UrlApiConfig

	def tearDown(self):
		pass

	# --------------------------------------------------------------
	def test_bad_auth(self):
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				authentication = 4
			Config()

	# --------------------------------------------------------------
	def test_not_init_auth(self):
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				authentication = UserPassAuthConfig
			Config()

	# --------------------------------------------------------------
	def test_good_auth(self):
		class Config(self.UrlApiConfig):
			authentication = UserPassAuthConfig()
		Config()

class TestMainConfiguration(unittest.TestCase):

	UrlApiConfig = None

	def setUp(self):
		class UrlApiConfig(APIConfig):
			ep = ResourceConfig(path=[''], method='GET')
		self.UrlApiConfig = UrlApiConfig

	def tearDown(self):
		pass

	# --------------------------------------------------------------
	def test_bad_verify_ssl(self):
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				url = 'http://localhost'
				verify_ssl = 3
			Config()

	# --------------------------------------------------------------
	def test_good_verify_ssl(self):
		class Config(self.UrlApiConfig):
			url = 'http://localhost'
			verify_ssl = True
		Config()

	# --------------------------------------------------------------
	def test_bad_server(self):
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				url = 'http://badserver'
			Config()

	# --------------------------------------------------------------
	def test_good_server(self):
		class Config(self.UrlApiConfig):
			url = 'http://localhost'
		Config()

		class Config(self.UrlApiConfig):
			url = 'http://server.tld'
		Config()

		class Config(self.UrlApiConfig):
			url = 'http://server.tld:8080'
		Config()


class TestResourceClass(unittest.TestCase):

	UrlApiConfig = None

	def setUp(self):
		class UrlApiConfig(APIConfig):
			url = 'http://localhost'
		self.UrlApiConfig = UrlApiConfig

	def tearDown(self):
		pass

	# --------------------------------------------------------------
	def test_empty_processor(self):
		class Config(self.UrlApiConfig):
			ep = ResourceConfig(path=[''], method='GET',
						  processor=JSONResource()
                       )
		Config()

	# --------------------------------------------------------------
	def test_good_processor(self):
		class Config(self.UrlApiConfig):
			ep = ResourceConfig(path=[''], method='GET',
						  processor=JSONResource(extract_section=['a', 'b', 'c'], create_attribute='sink')
                       )
		Config()

	# --------------------------------------------------------------
	def test_bad_processor(self):
		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				ep = ResourceConfig(path=[''], method='GET',
							  processor='error'
                        )

		with self.assertRaises(RestClientConfigurationError):
			class Config(self.UrlApiConfig):
				ep = ResourceConfig(path=[''], method='GET',
							  processor=JSONResource
                        )

		with self.assertRaises(TypeError):
			class Config(self.UrlApiConfig):
				ep = ResourceConfig(path=[''], method='GET',
							  processor=JSONResource('error')
                        )




"""The setuptools based setup module for rest-client.

Following the guidelines at:
https://packaging.python.org/en/latest/distributing.html

Used as a starting point:
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# get version from init
for line in open('rest_client/__init__.py'):
    if (line.startswith('__version__')):
        exec(line.strip())
        break


# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
#    name='universal-rest-client',
    name='qrest',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    #version='0.9.0',
    version=__version__, 

    description='A generic python REST client',
    long_description=long_description,

    # The project's main homepage.
    url='https://bitbucket.org/nhm_bioinformatics/qrest/',
    
    # Author details
    author='Joris Benschop (BASF Vegetable Seeds)',
    author_email='joris.benschop@gmail.com',
    maintainer='Joris Benschop', 
    maintainer_email='joris.benschop@gmail.com', 

    # Choose your license
    license='GPLv3',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',
        #'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Information Technology',
        'Topic :: Software Development :: Libraries :: Python Modules',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    # What does your project relate to?
    keywords='generic REST API client',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['docs', 'test']),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #py_modules=["rest_client"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['requests', 'uritools'],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': ['Sphinx'],
        'test': ['requests-mock'],
    },
)



from pprint import pprint
import unittest

import rest_client
from rest_client import APIConfig, RestClientConfigurationError
from rest_client import QueryParameter, BodyParameter, ResourceConfig
from rest_client import JSONResource
from rest_client.auth import UserPassAuthConfig


class JsonPlaceHolderConfig(APIConfig):
	url = 'https://jsonplaceholder.typicode.com'
	default_headers = {"Content-type": "application/json; charset=UTF-8"}
	
	all_posts = ResourceConfig(path=['posts'],
							   method='GET',
							   description='retrieve all posts'
                            )

	filter_posts = ResourceConfig(path=['posts'],
							   method='GET',
							   description='retrieve all posts by filter criteria',
							   parameters={
								   'user_id': QueryParameter(name='userId',
														  required=False,
														  # default='',
														  # choices=[],
														  description='the user Id that made the post'),
            }
        )

	single_post = ResourceConfig(path=['posts', '{item}'],
								 method='GET',
								 description='Retrieve a single post',
								 path_description={'item': 'select the post ID to retrieve',
                           },
            headers={'X-test-post': 'qREST python ORM', }
        )

	comments = ResourceConfig(path=['posts', '{post_id}', 'comments'],
								 method='GET',
								 description='Retrieve comments for a single post',
								 path_description={'post_id': 'select the post ID to retrieve', }
                           )

	create_post = ResourceConfig(path=['posts', ],
								 method='POST',
								 description='Create a new post',
								 parameters={
									 'title': BodyParameter(name='title',
															required=True,
															# default='',
															# choices=[],
															description='The title of the post'),
  									 'content': BodyParameter(name='body',
															  required=True,
															  # default='',
															  # choices=[],
															  description='The title of the post'),
  									 'user_id': BodyParameter(name='userId',
															  required=False,
															  default=101,
															  # choices=[],
															  description='The id of the user creating the post'),
        }
        )

class TestJsonPlaceHolderGet(unittest.TestCase):

	def setUp(self):
		self.config = JsonPlaceHolderConfig()
		x = 1

	def test_get_all_posts(self):
		'''
		check method or path
		'''
		x = rest_client.API(self.config)
		response = x.all_posts.fetch()
		self.assertIsInstance(response, list)
		self.assertEqual(len(response), 100)

	def test_get_post_1(self):
		'''
		check method or path
		'''
		x = rest_client.API(self.config)
		response = x.single_post.fetch(item=1)
		self.assertIsInstance(response, dict)
		self.assertEqual(len(response), 4)

	def test_filter_posts_fetch(self):
		'''
		check method or path
		'''
		x = rest_client.API(self.config)
		response = x.filter_posts.fetch(user_id=1)
		self.assertIsInstance(response, list)
		self.assertEqual(len(response), 10)

	def test_filter_posts_response(self):
		'''
		check method or path
		'''
		x = rest_client.API(self.config)
		resource = x.filter_posts(user_id=1)
		response = resource.data

		self.assertIsInstance(response, list)
		self.assertEqual(len(response), 10)

	def test_posts_comments(self):
		'''
		check method or path
		'''
		x = rest_client.API(self.config)
		response = x.comments.fetch(post_id=1)
		self.assertIsInstance(response, list)
		self.assertEqual(len(response), 5)



class TestJsonPlaceHolderPost(unittest.TestCase):

	def setUp(self):
		self.config = JsonPlaceHolderConfig()
		x = 1

	def test_new_post(self):
		'''
		create a new post
		'''
		x = rest_client.API(self.config)
		print(x.create_post.help('title'))

		title = "new post using qREST ORM"
		content = 'this is the new data posted using qREST'
		user_id = 200

		# send data to jsonplaceholder server
		response = x.create_post(title=title, content=content, user_id=user_id)
		self.assertIsInstance(response.data, dict)

		self.assertEqual(response.data['userId'], user_id)

		# also test using default values
		response = x.create_post(title=title, content=content)
		self.assertIsInstance(response.data, dict)

		self.assertEqual(response.data['userId'], 101)
		



'''
The main working module of the package: contains the API class (which creates the ORM for a
RES server), and the Resource class (which wraps around a single endpoint / resources ). The Resource
class is encouraged to be subclassed to add functionality such as complex pagination or response processing

'''

import importlib
import requests
import logging
from urllib.parse import quote, urljoin
from abc import ABC, abstractmethod
from typing import Optional, Type

from requests.packages.urllib3 import disable_warnings
from requests.packages.urllib3.exceptions import InsecureRequestWarning
disable_warnings(InsecureRequestWarning)

# ================================================================================================
# local imports
from .utils import URLValidator
from .exception import RestClientQueryError, RestClientConfigurationError, RestLoginError, RestResourceHTTPError, InvalidResourceError
from .response import JSONResponse

logger = logging.getLogger(__name__)

# ================================================================================================
class API:
    '''
    This is the main point of contact for end users
    '''

    # placeholder for subclassed resources
    config = None
    auth = None

    def __init__(self, config):
        """
        REST Client constructor
        
        :param config: The configuration object of the REST API resources
        :type config: Subclass of APIConfig
        
        An API describes a REST server, and contains a list of resources. 
        We allow customization of the "resource class", which is basically
        a wrapper around the response object: default is JSON, so we pre-made a JSON resource class.
        Optionally this resource class handles non-standard responses such as pagination or a
        specific response format from which the payload needs to be derived
        """

        # check
        from .conf import APIConfig
        if not isinstance(config, APIConfig):
            raise RestClientConfigurationError('configuration is not a APIConfig instance')


        self.config = config
        self.verifySSL = config.verify_ssl
        self.auth = config.authentication

        #  process the endpoints
        for name, item_config in self.config.endpoints.items():
            if not isinstance(item_config.processor, Resource):
                raise RestClientConfigurationError(f'defined resource class for {ep} is not a Resource instance')
            new_resource = self._create_rest_resource(item_config.processor,
                                                      resource_name=name,
                                                      config=item_config)
            setattr(self, name, new_resource)
            xx = 1
        x = 1


    # ---------------------------------------------------------------------------------------------
    @property
    def resources(self):
        """ Lists the available resources for this REST API

            :return: A list of the available resources for this REST API
            :rtype: ``list(string_type)``
        """

        resources = []
        fieldnames = dir(self)
        for fieldname in fieldnames:
            # this exclusion is to prevent endless loops
            if fieldname == 'resources':
                continue
            field = getattr(self, fieldname)
            if isinstance(field, Resource):
                resources.append(field.name)
        return resources


    # ---------------------------------------------------------------------------------------------
    def _create_rest_resource(self, processor, resource_name, config):
        """ This function is used to dynamically create request functions for a specified REST API resource

            :param resource: A string that represents the REST API resource
            :type resource: ``string_type``

            :return: A function that builds and sends a request for the specified REST API resource and validates the function call
            :rtype: ``list(string_type)``
        """

        if not config:
            raise InvalidResourceError(name=type(self).__name__, resource=resource_name)

        if not isinstance(processor, Resource):
            raise RestClientConfigurationError('processor must be a class instance')

        processor.configure(name=resource_name, config=config, server_url=self.config.url)
        return processor


# ===================================================================================================
class Resource(ABC):
    '''
    A resource is defined as a single REST endpoint.
    This class wraps functionality of creating and querying this resources, starting with a
    configuration string
    '''

    is_configured = False
    config = None

    server_url = None
    request_parameters = None
    verify_ssl = False
    auth = None
    cleaned_data = None
    
    response_class = None

    @abstractmethod
    def __init__(self):
        ''' ABC to set response parameters during configuration
        '''

    # ---------------------------------------------------------------------------------------------
    def configure(self,
                  name: str,
                  server_url: str,
                  config,
                  auth=None,
                  verify_ssl: bool = False
                  ):
        ''' Configure the resource. This is a required procedure to set all parameters.
        Setting these parameters is not possible by using __init__, because this class is initialized
        within the config, to enable setting custom parameters instead
        
        :param name: the pythonic name of this resource (i.e. my own name). Used to generate error messages
        :param server_url: the base server URL (e.g. http://localhost:8080)
        :param verify_ssl: boolean to set verify_ssl in the request on or off
        :param auth: Which Authentication module to use
        :type auth: subclass of AuthConfig 
        :param config: which ResourceConfiguration to use
        :type config: subclass of ResourceConfig 
        '''

        self.name = name
        self.server_url = server_url
        self.config = config
        self.auth = auth
        self.verify_ssl = verify_ssl
        
        self.cleaned_data = {}
        self.request_parameters = None
        self.is_configured = True
        

    # ---------------------------------------------------------------------------------------------
    def __call__(self, *args, **kwargs):
        '''
        runs the REST query using supplied arguments, checks input quality and format the REST
        parameters.
        '''

        self.cleaned_data = {}
        self.check(**kwargs)
        return self._get()


    # ---------------------------------------------------------------------------------------------
    def fetch(self, *args, **kwargs):
        '''
        shortcut function to immedaitely deliver the content of the response insetead of the response object itself
        '''
        response = self.__call__(*args, **kwargs)
        return response.fetch()

    # ---------------------------------------------------------------------------------------------
    @property
    def parameters(self) -> dict:
        '''
        return the configuration parameters for this rest resource
        :return: A dictionary of the 'optional', 'required' and 'multiple' (keys) query parameters (value, a list)

        '''
        return self.config.as_dict

    #---------------------------------------------------------------------------------------------
    @property
    def description(self) -> str:
        '''
        shortcut to provide a description of the endpoint
        '''
        return self.help()


    def help(self, parameter_name: Optional[str]=None):
        '''
        Print description of the endpoint and the parameters
        :param parameter_name: Optional parameter name to request the help text of
        '''
        if not parameter_name:
            return (self.config.description or 'No description given for this endpoint')
        if parameter_name not in self.config.all_parameters:
            if self.config.all_parameters:
                return '%s is not a valid parameter: valid are: %s' % (parameter_name, ','.join(self.config.all_parameters))
            else:
                return 'this endpoint has no parameters'
        if parameter_name in self.config.parameters:
            param_help = self.config.parameters[parameter_name].description
            choices = self.config.parameters[parameter_name].choices
            if choices:
                param_help += '. Valid choices are: %s' % ', '.join(choices)
            return (param_help or 'no description given for this parameter')
        if parameter_name in self.config.path_parameters:
            param_help = self.config.path_description.get(parameter_name, 'no description given for this parameter')
            return param_help
        return 'ERROR: not yet implemented'


    #---------------------------------------------------------------------------------------------
    def check(self, **kwargs):
        '''
        check the input request parameters before sending it to the remote service
        '''

        conf = self.config

        #----------------------------------
        # deny superfluous input
        diff = list(set(kwargs.keys()).difference(conf.all_parameters))
        if diff:
            raise RestClientQueryError("parameters {difference} are supplied but not usable for resource '{resource}'".format(
                difference=diff,
                resource=self.name
            ))

        #----------------------------------
        # Check required parameters
        for parameter in conf.required_parameters:
            if parameter not in kwargs:
                raise RestClientQueryError("parameter '{parameter}' is missing or empty for resource '{resource}'".format(
                    parameter=parameter,
                    resource=self.name
                ))

        #----------------------------------
        # check choices
        for parameter in kwargs:
            if parameter not in self.config.parameters:
                continue
            config = self.config.parameters[parameter]
            if config.choices:
                if not kwargs[parameter] in config.choices:
                    raise RestClientQueryError("value '{val}' for parameter '{parameter}' is not a valid choice: pick from {choices}".format(
                        val = kwargs[parameter],
                        parameter=parameter,
                        choices=', '.join(config.choices))
                                               )

        #----------------------------------
        # check query parameters
        intersection = set(conf.all_query_parameters).intersection(kwargs.keys())
        groups_used = {}
        for kwarg in intersection:
            for group in conf.query_parameter_groups:
                if kwarg in conf.query_parameter_groups[group]:
                    if group in groups_used:
                        raise RestClientQueryError(
                            "parameter '{kwarg1}' and '{kwarg2}' from group '{group}' can't be used together".format(
                                kwarg1=kwarg,
                                kwarg2=groups_used[group],
                                group=group
                            ))
                    else:
                        groups_used[group] = kwarg
                    break
            if isinstance(kwargs[kwarg], list) and kwarg not in conf.multiple_parameters:
                raise RestClientQueryError("parameter '{kwarg}' is not multiple".format(
                    kwarg=kwarg
                ))


        # apply defaults for missing optional parameters that do have default values
        defaults = self.config.defaults
        for item, value in defaults.items():
            if item not in kwargs:
                kwargs[item] = value


        self.cleaned_data = kwargs

    #---------------------------------------------------------------------------------------------
    @property
    def query_url(self):
        '''
        returns the URL that is actually queried
        '''

        # url and parameters
        if 'cleaned_data' not in dir(self):
            raise KeyError('request data is not cleaned. Run validate_request first')


        resolved_path = "/".join(self.config.path)
        selected_params = [parameter for parameter in self.cleaned_data if parameter in self.config.path_parameters]
        path_para = {p: quote(str(self.cleaned_data[p]), safe='') for p in selected_params}
        resolved_path = resolved_path.format(**path_para)

        # Construct URL using base URL and path
        url = urljoin(base=self.server_url, url=resolved_path)

        # Check if valid URL
        # Only allow http or https schemes for the REST API base URL
        url_validator = URLValidator(schemes=["http", "https"])
        url_validator.check(url)

        return url

    # ---------------------------------------------------------------------------------------------
    @property
    def query_parameters(self):
        '''
        generate the request and body parameters based on the validated input and the config
        '''
        request_parameters = {}
        body_parameters = {}

        # process via the config
        config_parameters = self.config.parameters
        for para_name, para_val in self.cleaned_data.items():
            if para_name in self.config.path_parameters:
                continue
            rest_name = config_parameters[para_name].name
            if config_parameters[para_name].call_location == 'query':
                request_parameters[rest_name] = para_val
            elif config_parameters[para_name].call_location == 'body':
                body_parameters[rest_name] = para_val
            else:
                raise RestClientConfigurationError('call location for %s is not understood' % para_name)

        # collect and return
        return_structure = {'request': request_parameters,
                            'body': body_parameters,
                            }

        return return_structure

    # ---------------------------------------------------------------------------------------------
    def _get(self, extra_request=None, extra_body=None):
        """ This function builds and sends a request for a specified REST API resource.
            The parameters are validated in a previous call to validate_query().
            It returns a dictionary of the response or throws an appropriate
            error, depending on the HTTP return code.
            
            This should be the *only* place in the module where the Requests module is called!
            
        """

        # check if user is logged in
        if self.auth and not self.auth.is_logged_in:
            raise RestLoginError('user is not logged in')

        # url and parameters
        if 'cleaned_data' not in dir(self):
            raise KeyError('request data is not cleaned. Run validate_query first')

        query_parameters = self.query_parameters

        # add hooks to extend get function
        for location, data_dict in [('request', extra_request), ('body', extra_body)]:
            if not data_dict:
                continue
            if not isinstance(data_dict, dict):
                raise RestClientQueryError('extra_request and extra_body must be dict')
            for item in data_dict.keys():
                if item in query_parameters[location]:
                    raise RestClientQueryError('trying to overload parameter ' + item)
            query_parameters[location].update(data_dict)

        # Do HTTP request to REST API
        logger.debug('[RESTCLIENT]: running %s' % self.query_url)
        try:
            response = requests.request(method=self.config.method,
                                        auth=self.auth,
                                        verify=self.verify_ssl,
                                        url=self.query_url,
                                        params=query_parameters['request'],
                                        json=query_parameters['body'],
                                        headers=self.config.headers
                                        )
            assert isinstance(response, requests.Response)

            if response.status_code > 399:  #Nicely catch exceptions
                raise RestResourceHTTPError(response_object=response)
            # for completeness sake: let requests check for valid output
            # code should not get here...
            response.raise_for_status()
        except ValueError:
            # Weird response errors: just give back the raw data. This has the risk of dismissing
            # valid errors!
            return response.content
        except requests.HTTPError as http:
            # This is a back-catcher for HTTP errors that were not caught before. Code shoul
            # not get here
            raise http
        else:
            r = self.response_class(response)
            return r


# ###############################################################
class JSONResource(Resource):
    """ A REST Resource that expects a JSON return
    
    """

    def __init__(self, *, extract_section: Optional[list] = None, create_attribute: Optional[str] = 'results'):
        '''
        :param extract_section: This indicates which part of the obtained JSON response contains the main
            payload that should be extracted. The tree is provided as a list of items to traverse
        :param create_attribute: The "results_name" which is the property that will be generated to contain
            the previously obtained subsection of the json tree
        '''

        self.response_class = JSONResponse(extract_section, create_attribute)


